# HG changeset patch
# User erankor <erankor@gmail.com>
# Date 1502100634 14400
#      Mon Aug 07 06:10:34 2017 -0400
# Node ID cb27bda9557fede50b531a0b3b4db98b38cc937a
# Parent  ac120e797d284a1ef469e44a379c2aafeb0747a0
Log: support buffering in scripted logs.

Added support for buffering in logs that use variables in the file name.
The feature relies on the use of open file cache - an ngx_open_file_t is stored
along with each ngx_open_file_cache_t. The ngx_open_file_t holds the buffer and
flush event.

diff -r ac120e797d28 -r cb27bda9557f src/core/ngx_open_file_cache.c
--- a/src/core/ngx_open_file_cache.c	Fri Aug 04 17:03:10 2017 +0300
+++ b/src/core/ngx_open_file_cache.c	Mon Aug 07 06:10:34 2017 -0400
@@ -42,6 +42,8 @@
 static void ngx_open_file_cleanup(void *data);
 static void ngx_close_cached_file(ngx_open_file_cache_t *cache,
     ngx_cached_open_file_t *file, ngx_uint_t min_uses, ngx_log_t *log);
+static void ngx_free_cached_file(ngx_open_file_cache_t *cache,
+    ngx_cached_open_file_t *file);
 static void ngx_open_file_del_event(ngx_cached_open_file_t *file);
 static void ngx_expire_old_cached_files(ngx_open_file_cache_t *cache,
     ngx_uint_t n, ngx_log_t *log);
@@ -72,6 +74,7 @@
     cache->current = 0;
     cache->max = max;
     cache->inactive = inactive;
+    ngx_memzero(&cache->ext, sizeof(cache->ext));
 
     cln = ngx_pool_cleanup_add(pool, 0);
     if (cln == NULL) {
@@ -121,8 +124,7 @@
             ngx_close_cached_file(cache, file, 0, ngx_cycle->log);
 
         } else {
-            ngx_free(file->name);
-            ngx_free(file);
+            ngx_free_cached_file(cache, file);
         }
     }
 
@@ -332,6 +334,10 @@
 
             ngx_open_file_del_event(file);
 
+            if (cache->ext.flush) {
+                cache->ext.flush(file + 1);
+            }
+
             if (ngx_close_file(file->fd) == NGX_FILE_ERROR) {
                 ngx_log_error(NGX_LOG_ALERT, pool->log, ngx_errno,
                               ngx_close_file_n " \"%V\" failed", name);
@@ -363,7 +369,8 @@
         ngx_expire_old_cached_files(cache, 0, pool->log);
     }
 
-    file = ngx_alloc(sizeof(ngx_cached_open_file_t), pool->log);
+    file = ngx_alloc(sizeof(ngx_cached_open_file_t) + cache->ext.size,
+                     pool->log);
 
     if (file == NULL) {
         goto failed;
@@ -377,6 +384,15 @@
         goto failed;
     }
 
+    if (cache->ext.init) {
+        if (cache->ext.init(file + 1, of->ctx) != NGX_OK) {
+            ngx_free(file->name);
+            ngx_free(file);
+            file = NULL;
+            goto failed;
+        }
+    }
+
     ngx_cpystrn(file->name, name->data, name->len + 1);
 
     file->node.key = hash;
@@ -445,6 +461,8 @@
             ofcln->log = pool->log;
         }
 
+        of->ctx = file + 1;
+
         return NGX_OK;
     }
 
@@ -460,6 +478,10 @@
         if (file->count == 0) {
 
             if (file->fd != NGX_INVALID_FILE) {
+                if (cache->ext.flush) {
+                    cache->ext.flush(file + 1);
+                }
+
                 if (ngx_close_file(file->fd) == NGX_FILE_ERROR) {
                     ngx_log_error(NGX_LOG_ALERT, pool->log, ngx_errno,
                                   ngx_close_file_n " \"%s\" failed",
@@ -467,8 +489,7 @@
                 }
             }
 
-            ngx_free(file->name);
-            ngx_free(file);
+            ngx_free_cached_file(cache, file);
 
         } else {
             file->close = 1;
@@ -1057,6 +1078,10 @@
 
     if (file->fd != NGX_INVALID_FILE) {
 
+        if (cache->ext.flush) {
+            cache->ext.flush(file + 1);
+        }
+
         if (ngx_close_file(file->fd) == NGX_FILE_ERROR) {
             ngx_log_error(NGX_LOG_ALERT, log, ngx_errno,
                           ngx_close_file_n " \"%s\" failed", file->name);
@@ -1069,6 +1094,18 @@
         return;
     }
 
+    ngx_free_cached_file(cache, file);
+}
+
+
+static void
+ngx_free_cached_file(ngx_open_file_cache_t *cache,
+    ngx_cached_open_file_t *file)
+{
+    if (cache->ext.free) {
+        cache->ext.free(file + 1);
+    }
+
     ngx_free(file->name);
     ngx_free(file);
 }
@@ -1135,8 +1172,7 @@
             ngx_close_cached_file(cache, file, 0, log);
 
         } else {
-            ngx_free(file->name);
-            ngx_free(file);
+            ngx_free_cached_file(cache, file);
         }
     }
 }
diff -r ac120e797d28 -r cb27bda9557f src/core/ngx_open_file_cache.h
--- a/src/core/ngx_open_file_cache.h	Fri Aug 04 17:03:10 2017 +0300
+++ b/src/core/ngx_open_file_cache.h	Mon Aug 07 06:10:34 2017 -0400
@@ -17,6 +17,7 @@
 
 
 typedef struct {
+    void                    *ctx;
     ngx_fd_t                 fd;
     ngx_file_uniq_t          uniq;
     time_t                   mtime;
@@ -89,6 +90,14 @@
 
 
 typedef struct {
+    size_t                   size;
+    ngx_int_t              (*init)(void *ctx, void *init_ctx);
+    void                   (*flush)(void *ctx);
+    void                   (*free)(void *ctx);
+} ngx_open_file_cache_ext_t;
+
+
+typedef struct {
     ngx_rbtree_t             rbtree;
     ngx_rbtree_node_t        sentinel;
     ngx_queue_t              expire_queue;
@@ -96,6 +105,8 @@
     ngx_uint_t               current;
     ngx_uint_t               max;
     time_t                   inactive;
+
+    ngx_open_file_cache_ext_t ext;
 } ngx_open_file_cache_t;
 
 
diff -r ac120e797d28 -r cb27bda9557f src/http/modules/ngx_http_log_module.c
--- a/src/http/modules/ngx_http_log_module.c	Fri Aug 04 17:03:10 2017 +0300
+++ b/src/http/modules/ngx_http_log_module.c	Mon Aug 07 06:10:34 2017 -0400
@@ -56,16 +56,29 @@
 
 
 typedef struct {
+    ssize_t                     size;
+    ngx_int_t                   gzip;
+    ngx_msec_t                  flush;
+} ngx_http_log_buf_params_t;
+
+typedef struct {
     ngx_array_t                *lengths;
     ngx_array_t                *values;
 } ngx_http_log_script_t;
 
 
 typedef struct {
+    ngx_http_log_buf_params_t  *buf_params;
+    ngx_log_t                  *log;
+} ngx_http_log_cached_file_init_ctx_t;
+
+
+typedef struct {
     ngx_open_file_t            *file;
     ngx_http_log_script_t      *script;
     time_t                      disk_full_time;
     time_t                      error_log_time;
+    ngx_http_log_buf_params_t   buf_params;
     ngx_syslog_peer_t          *syslog_peer;
     ngx_http_log_fmt_t         *format;
     ngx_http_complex_value_t   *filter;
@@ -91,9 +104,9 @@
 
 
 static void ngx_http_log_write(ngx_http_request_t *r, ngx_http_log_t *log,
-    u_char *buf, size_t len);
-static ssize_t ngx_http_log_script_write(ngx_http_request_t *r,
-    ngx_http_log_script_t *script, u_char **name, u_char *buf, size_t len);
+    ngx_open_file_t *file, u_char *buf, size_t len);
+static ngx_open_file_t *ngx_http_log_script_run(
+    ngx_http_request_t *r, ngx_http_log_t *log);
 
 #if (NGX_ZLIB)
 static ssize_t ngx_http_log_gzip(ngx_fd_t fd, u_char *buf, size_t len,
@@ -250,6 +263,7 @@
     ngx_str_t                 val;
     ngx_uint_t                i, l;
     ngx_http_log_t           *log;
+    ngx_open_file_t          *file;
     ngx_http_log_op_t        *op;
     ngx_http_log_buf_t       *buffer;
     ngx_http_log_loc_conf_t  *lcf;
@@ -287,6 +301,17 @@
             continue;
         }
 
+        if (log->script) {
+            file = ngx_http_log_script_run(r, &log[l]);
+            if (file == NULL) {
+                continue;
+            }
+
+        }
+        else {
+            file = log[l].file;
+        }
+
         ngx_http_script_flush_no_cacheable_variables(r, log[l].format->flushes);
 
         len = 0;
@@ -312,13 +337,13 @@
 
         len += NGX_LINEFEED_SIZE;
 
-        buffer = log[l].file ? log[l].file->data : NULL;
+        buffer = file->data;
 
         if (buffer) {
 
             if (len > (size_t) (buffer->last - buffer->pos)) {
 
-                ngx_http_log_write(r, &log[l], buffer->start,
+                ngx_http_log_write(r, &log[l], file, buffer->start,
                                    buffer->pos - buffer->start);
 
                 buffer->pos = buffer->start;
@@ -386,7 +411,7 @@
 
         ngx_linefeed(p);
 
-        ngx_http_log_write(r, &log[l], line, p - line);
+        ngx_http_log_write(r, &log[l], file, line, p - line);
     }
 
     return NGX_OK;
@@ -394,10 +419,9 @@
 
 
 static void
-ngx_http_log_write(ngx_http_request_t *r, ngx_http_log_t *log, u_char *buf,
-    size_t len)
+ngx_http_log_write(ngx_http_request_t *r, ngx_http_log_t *log,
+    ngx_open_file_t *file, u_char *buf,  size_t len)
 {
-    u_char              *name;
     time_t               now;
     ssize_t              n;
     ngx_err_t            err;
@@ -405,27 +429,19 @@
     ngx_http_log_buf_t  *buffer;
 #endif
 
-    if (log->script == NULL) {
-        name = log->file->name.data;
-
 #if (NGX_ZLIB)
-        buffer = log->file->data;
-
-        if (buffer && buffer->gzip) {
-            n = ngx_http_log_gzip(log->file->fd, buf, len, buffer->gzip,
-                                  r->connection->log);
-        } else {
-            n = ngx_write_fd(log->file->fd, buf, len);
-        }
+    buffer = file->data;
+
+    if (buffer && buffer->gzip) {
+        n = ngx_http_log_gzip(file->fd, buf, len, buffer->gzip,
+                                r->connection->log);
+    } else {
+        n = ngx_write_fd(file->fd, buf, len);
+    }
 #else
-        n = ngx_write_fd(log->file->fd, buf, len);
+    n = ngx_write_fd(file->fd, buf, len);
 #endif
 
-    } else {
-        name = NULL;
-        n = ngx_http_log_script_write(r, log->script, &name, buf, len);
-    }
-
     if (n == (ssize_t) len) {
         return;
     }
@@ -441,7 +457,7 @@
 
         if (now - log->error_log_time > 59) {
             ngx_log_error(NGX_LOG_ALERT, r->connection->log, err,
-                          ngx_write_fd_n " to \"%s\" failed", name);
+                          ngx_write_fd_n " to \"%s\" failed", file->name.data);
 
             log->error_log_time = now;
         }
@@ -452,23 +468,117 @@
     if (now - log->error_log_time > 59) {
         ngx_log_error(NGX_LOG_ALERT, r->connection->log, 0,
                       ngx_write_fd_n " to \"%s\" was incomplete: %z of %uz",
-                      name, n, len);
+                      file->name.data, n, len);
 
         log->error_log_time = now;
     }
 }
 
 
-static ssize_t
-ngx_http_log_script_write(ngx_http_request_t *r, ngx_http_log_script_t *script,
-    u_char **name, u_char *buf, size_t len)
+static ngx_int_t
+ngx_http_log_cached_file_init(void *ctx, void *init_ctx)
 {
-    size_t                     root;
-    ssize_t                    n;
-    ngx_str_t                  log, path;
-    ngx_open_file_info_t       of;
-    ngx_http_log_loc_conf_t   *llcf;
-    ngx_http_core_loc_conf_t  *clcf;
+    ngx_http_log_cached_file_init_ctx_t *init = init_ctx;
+    ngx_http_log_buf_params_t           *params = init->buf_params;
+    ngx_open_file_t                     *file = ctx;
+    ngx_http_log_buf_t                  *buffer;
+
+    ngx_memzero(file, sizeof(*file));
+
+    if (params->size == 0) {
+        return NGX_OK;
+    }
+
+    buffer = ngx_alloc(sizeof(ngx_http_log_buf_t), init->log);
+    if (buffer == NULL) {
+        return NGX_ERROR;
+    }
+
+    buffer->start = ngx_alloc(params->size, init->log);
+    if (buffer->start == NULL) {
+        ngx_free(buffer);
+        return NGX_ERROR;
+    }
+
+    buffer->pos = buffer->start;
+    buffer->last = buffer->start + params->size;
+
+    if (params->flush) {
+        buffer->event = ngx_alloc(sizeof(ngx_event_t), init->log);
+        if (buffer->event == NULL) {
+            ngx_free(buffer->start);
+            ngx_free(buffer);
+            return NGX_ERROR;
+        }
+
+        ngx_memzero(buffer->event, sizeof(*buffer->event));
+        buffer->event->data = file;
+        buffer->event->handler = ngx_http_log_flush_handler;
+        buffer->event->log = ngx_cycle->log;
+        buffer->event->cancelable = 1;
+
+        buffer->flush = params->flush;
+
+    } else {
+        buffer->event = NULL;
+        buffer->flush = 0;
+    }
+
+    buffer->gzip = params->gzip;
+
+    file->data = buffer;
+
+    return NGX_OK;
+}
+
+
+static void
+ngx_http_log_cached_file_flush(void *ctx)
+{
+    ngx_open_file_t  *file = ctx;
+
+    if (file->data == NULL) {
+        return;
+    }
+
+    ngx_http_log_flush(file, ngx_cycle->log);
+}
+
+static void
+ngx_http_log_cached_file_free(void *ctx)
+{
+    ngx_open_file_t           *file = ctx;
+    ngx_http_log_buf_t        *buffer;
+
+    buffer = file->data;
+    if (buffer == NULL) {
+        return;
+    }
+
+    if (buffer->event) {
+        if (buffer->event->timer_set) {
+            ngx_del_timer(buffer->event);
+        }
+
+        ngx_free(buffer->event);
+    }
+
+    ngx_free(buffer->start);
+    ngx_free(buffer);
+}
+
+
+static ngx_open_file_t*
+ngx_http_log_script_run(ngx_http_request_t *r, ngx_http_log_t *log)
+{
+    size_t                              root;
+    ngx_str_t                           path;
+    ngx_open_file_t                    *file;
+    ngx_http_log_buf_t                 *buffer;
+    ngx_open_file_info_t                of;
+    ngx_http_log_loc_conf_t            *llcf;
+    ngx_http_core_loc_conf_t           *clcf;
+    ngx_http_log_cached_file_init_ctx_t init_ctx;
 
     clcf = ngx_http_get_module_loc_conf(r, ngx_http_core_module);
 
@@ -477,8 +587,7 @@
         /* test root directory existence */
 
         if (ngx_http_map_uri_to_path(r, &path, &root, 0) == NULL) {
-            /* simulate successful logging */
-            return len;
+            return NULL;
         }
 
         path.data[root] = '\0';
@@ -493,82 +602,115 @@
         of.events = clcf->open_file_cache_events;
 
         if (ngx_http_set_disable_symlinks(r, clcf, &path, &of) != NGX_OK) {
-            /* simulate successful logging */
-            return len;
+            return NULL;
         }
 
         if (ngx_open_cached_file(clcf->open_file_cache, &path, &of, r->pool)
             != NGX_OK)
         {
             if (of.err == 0) {
-                /* simulate successful logging */
-                return len;
+                return NULL;
             }
 
             ngx_log_error(NGX_LOG_ERR, r->connection->log, of.err,
                           "testing \"%s\" existence failed", path.data);
-
-            /* simulate successful logging */
-            return len;
+            return NULL;
         }
 
         if (!of.is_dir) {
             ngx_log_error(NGX_LOG_ERR, r->connection->log, NGX_ENOTDIR,
                           "testing \"%s\" existence failed", path.data);
-
-            /* simulate successful logging */
-            return len;
+            return NULL;
         }
     }
 
-    if (ngx_http_script_run(r, &log, script->lengths->elts, 1,
-                            script->values->elts)
+    if (ngx_http_script_run(r, &path, log->script->lengths->elts, 1,
+                            log->script->values->elts)
         == NULL)
     {
-        /* simulate successful logging */
-        return len;
+        return NULL;
     }
 
-    log.data[log.len - 1] = '\0';
-    *name = log.data;
+    path.data[path.len - 1] = '\0';
 
     ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
-                   "http log \"%s\"", log.data);
+                   "http log \"%s\"", path.data);
 
     llcf = ngx_http_get_module_loc_conf(r, ngx_http_log_module);
 
     ngx_memzero(&of, sizeof(ngx_open_file_info_t));
 
+    init_ctx.buf_params = &log->buf_params;
+    init_ctx.log = r->connection->log;
+
     of.log = 1;
     of.valid = llcf->open_file_cache_valid;
     of.min_uses = llcf->open_file_cache_min_uses;
     of.directio = NGX_OPEN_FILE_DIRECTIO_OFF;
-
-    if (ngx_http_set_disable_symlinks(r, clcf, &log, &of) != NGX_OK) {
-        /* simulate successful logging */
-        return len;
+    of.ctx = &init_ctx;
+
+    if (ngx_http_set_disable_symlinks(r, clcf, &path, &of) != NGX_OK) {
+        return NULL;
     }
 
-    if (ngx_open_cached_file(llcf->open_file_cache, &log, &of, r->pool)
+    if (ngx_open_cached_file(llcf->open_file_cache, &path, &of, r->pool)
         != NGX_OK)
     {
         if (of.err == 0) {
-            /* simulate successful logging */
-            return len;
+            return NULL;
         }
 
         ngx_log_error(NGX_LOG_CRIT, r->connection->log, ngx_errno,
-                      "%s \"%s\" failed", of.failed, log.data);
-        /* simulate successful logging */
-        return len;
+                      "%s \"%s\" failed", of.failed, path.data);
+        return NULL;
     }
 
     ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                    "http log #%d", of.fd);
 
-    n = ngx_write_fd(of.fd, buf, len);
-
-    return n;
+    if (of.ctx) {
+        file = of.ctx;
+
+        buffer = file->data;
+        if (buffer == NULL) {
+            if (log->buf_params.size) {
+                ngx_log_error(NGX_LOG_CRIT, r->connection->log, 0,
+                    "log file \"%V\" already opened "
+                    "without buffering",
+                    &path);
+                return NULL;
+            }
+
+        } else {
+            if (buffer->last - buffer->start != log->buf_params.size
+                || buffer->flush != log->buf_params.flush
+                || buffer->gzip != log->buf_params.gzip)
+            {
+                ngx_log_error(NGX_LOG_CRIT, r->connection->log, 0,
+                    "log file \"%V\" already opened "
+                    "with conflicting parameters",
+                    &path);
+                return NULL;
+            }
+        }
+
+    } else {
+        if (log->buf_params.size) {
+            ngx_log_error(NGX_LOG_CRIT, r->connection->log, 0,
+                "buffered logs with variables in name must use cache");
+            return NULL;
+        }
+
+        file = ngx_pcalloc(r->pool, sizeof(*file));
+        if (file == NULL) {
+            return NULL;
+        }
+    }
+
+    file->fd = of.fd;
+    file->name = path;
+
+    return file;
 }
 
 
@@ -1418,18 +1560,19 @@
 
     if (size) {
 
-        if (log->script) {
-            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
-                               "buffered logs cannot have variables in name");
-            return NGX_CONF_ERROR;
-        }
-
         if (log->syslog_peer) {
             ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                                "logs to syslog cannot be buffered");
             return NGX_CONF_ERROR;
         }
 
+        if (log->script) {
+            log->buf_params.size = size;
+            log->buf_params.flush = flush;
+            log->buf_params.gzip = gzip;
+            return NGX_CONF_OK;
+        }
+
         if (log->file->data) {
             buffer = log->file->data;
 
@@ -1707,7 +1850,8 @@
 static char *
 ngx_http_log_open_file_cache(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
 {
-    ngx_http_log_loc_conf_t *llcf = conf;
+    ngx_http_log_loc_conf_t   *llcf = conf;
+    ngx_open_file_cache_ext_t *ext;
 
     time_t       inactive, valid;
     ngx_str_t   *value, s;
@@ -1802,6 +1946,12 @@
 
     if (llcf->open_file_cache) {
 
+        ext = &llcf->open_file_cache->ext;
+        ext->size = sizeof(ngx_open_file_t);
+        ext->init = ngx_http_log_cached_file_init;
+        ext->flush = ngx_http_log_cached_file_flush;
+        ext->free = ngx_http_log_cached_file_free;
+
         llcf->open_file_cache_valid = valid;
         llcf->open_file_cache_min_uses = min_uses;
 
